@@ -19,6 +19,9 @@
 #include "sql_table.h"                       // mysql_alter_table,
                                              // mysql_exchange_partition
 #include "sql_alter.h"
+#include "rpl_mi.h"
+#include "slave.h"
+#include "debug_sync.h"
 #include "wsrep_mysqld.h"
 
 Alter_info::Alter_info(const Alter_info &rhs, MEM_ROOT *mem_root)
@@ -253,50 +256,6 @@ Alter_info::algorithm(const THD *thd) const
 }
 
 
-uint Alter_info::check_vcol_field(Item_field *item) const
-{
-  /*
-    vcol->flags are modified in-place, so we'll need to reset them
-    if ALTER fails for any reason
-  */
-  if (item->field && !item->field->table->needs_reopen())
-    item->field->table->mark_table_for_reopen();
-
-  if (!item->field &&
-      ((item->db_name.length && !db.streq(item->db_name)) ||
-       (item->table_name.length && !table_name.streq(item->table_name))))
-  {
-    char *ptr= (char*)current_thd->alloc(item->db_name.length +
-                                         item->table_name.length +
-                                         item->field_name.length + 3);
-    strxmov(ptr, safe_str(item->db_name.str), item->db_name.length ? "." : "",
-            item->table_name.str, ".", item->field_name.str, NullS);
-    item->field_name.str= ptr;
-    return VCOL_IMPOSSIBLE;
-  }
-  for (Key &k: key_list)
-  {
-    if (k.type != Key::FOREIGN_KEY)
-      continue;
-    Foreign_key *fk= (Foreign_key*) &k;
-    if (fk->update_opt < FK_OPTION_CASCADE &&
-        fk->delete_opt < FK_OPTION_SET_NULL)
-      continue;
-    for (Key_part_spec& kp: fk->columns)
-    {
-      if (item->field_name.streq(kp.field_name))
-        return VCOL_NON_DETERMINISTIC;
-    }
-  }
-  for (Create_field &cf: create_list)
-  {
-    if (item->field_name.streq(cf.field_name))
-      return cf.vcol_info ? cf.vcol_info->flags : 0;
-  }
-  return 0;
-}
-
-
 Alter_table_ctx::Alter_table_ctx()
   : db(null_clex_str), table_name(null_clex_str), alias(null_clex_str),
     new_db(null_clex_str), new_name(null_clex_str), new_alias(null_clex_str)
@@ -452,7 +411,7 @@ bool Sql_cmd_alter_table::execute(THD *thd)
     referenced from this structure will be modified.
     @todo move these into constructor...
   */
-  HA_CREATE_INFO create_info(lex->create_info);
+  Table_specification_st create_info(lex->create_info);
   Alter_info alter_info(lex->alter_info, thd->mem_root);
   create_info.alter_info= &alter_info;
   privilege_t priv(NO_ACL);
@@ -471,6 +430,8 @@ bool Sql_cmd_alter_table::execute(THD *thd)
     as for RENAME TO, as being done by SQLCOM_RENAME_TABLE
   */
   if ((alter_info.partition_flags & ALTER_PARTITION_DROP) ||
+      (alter_info.partition_flags & ALTER_PARTITION_CONVERT_IN) ||
+      (alter_info.partition_flags & ALTER_PARTITION_CONVERT_OUT) ||
       (alter_info.flags & ALTER_RENAME))
     priv_needed|= DROP_ACL;
 
