@@ -251,14 +251,9 @@ static XID_cache_element *xid_cache_search(THD *thd, XID *xid)
                                         xid->key(), xid->key_length());
   if (element)
   {
-    /* The element can be removed from lf_hash by other thread, but
-    element->acquire_recovered() will return false in this case. */
     if (!element->acquire_recovered())
       element= 0;
     lf_hash_search_unpin(thd->xid_hash_pins);
-    /* Once the element is acquired (i.e. got the ACQUIRED bit) by this thread,
-    only this thread can delete it. The deletion happens in xid_cache_delete().
-    See also the XID_cache_element documentation. */
     DEBUG_SYNC(thd, "xa_after_search");
   }
   return element;
@@ -401,7 +396,7 @@ bool xa_trans_force_rollback(THD *thd)
     rc= true;
   }
   thd->variables.option_bits&=
-    ~(OPTION_BEGIN | OPTION_KEEP_LOG | OPTION_GTID_BEGIN);
+    ~(OPTION_BEGIN | OPTION_BINLOG_THIS_TRX | OPTION_GTID_BEGIN);
   thd->transaction->all.reset();
   thd->server_status&=
     ~(SERVER_STATUS_IN_TRANS | SERVER_STATUS_IN_TRANS_READONLY);
@@ -538,7 +533,7 @@ bool trans_xa_prepare(THD *thd)
     {
       if (!mdl_request.ticket)
         ha_rollback_trans(thd, TRUE);
-      thd->variables.option_bits&= ~(OPTION_BEGIN | OPTION_KEEP_LOG);
+      thd->variables.option_bits&= ~(OPTION_BEGIN | OPTION_BINLOG_THIS_TRX);
       thd->transaction->all.reset();
       thd->server_status&=
         ~(SERVER_STATUS_IN_TRANS | SERVER_STATUS_IN_TRANS_READONLY);
@@ -606,16 +601,6 @@ bool trans_xa_commit(THD *thd)
     if (auto xs= xid_cache_search(thd, thd->lex->xid))
     {
       bool xid_deleted= false;
-      MDL_request mdl_request;
-      bool rw_trans= (xs->rm_error != ER_XA_RBROLLBACK);
-
-      if (rw_trans && thd->is_read_only_ctx())
-      {
-        my_error(ER_OPTION_PREVENTS_STATEMENT, MYF(0), "--read-only");
-        res= 1;
-        goto _end_external_xid;
-      }
-
       res= xa_trans_rolled_back(xs);
       /*
         Acquire metadata lock which will ensure that COMMIT is blocked
@@ -624,6 +609,7 @@ bool trans_xa_commit(THD *thd)
 
         We allow FLUSHer to COMMIT; we assume FLUSHer knows what it does.
       */
+      MDL_request mdl_request;
       MDL_REQUEST_INIT(&mdl_request, MDL_key::BACKUP, "", "", MDL_BACKUP_COMMIT,
                        MDL_EXPLICIT);
       if (thd->mdl_context.acquire_lock(&mdl_request,
@@ -673,11 +659,7 @@ bool trans_xa_commit(THD *thd)
     DBUG_RETURN(res);
   }
 
-  if (thd->transaction->all.is_trx_read_write() && thd->is_read_only_ctx())
-  {
-    my_error(ER_OPTION_PREVENTS_STATEMENT, MYF(0), "--read-only");
-    DBUG_RETURN(TRUE);
-  } else if (xa_trans_rolled_back(xid_state.xid_cache_element))
+  if (xa_trans_rolled_back(xid_state.xid_cache_element))
   {
     xa_trans_force_rollback(thd);
     DBUG_RETURN(thd->is_error());
@@ -744,7 +726,7 @@ bool trans_xa_commit(THD *thd)
     DBUG_RETURN(TRUE);
   }
 
-  thd->variables.option_bits&= ~(OPTION_BEGIN | OPTION_KEEP_LOG);
+  thd->variables.option_bits&= ~(OPTION_BEGIN | OPTION_BINLOG_THIS_TRX);
   thd->transaction->all.reset();
   thd->server_status&=
     ~(SERVER_STATUS_IN_TRANS | SERVER_STATUS_IN_TRANS_READONLY);
@@ -795,15 +777,6 @@ bool trans_xa_rollback(THD *thd)
       bool res;
       bool xid_deleted= false;
       MDL_request mdl_request;
-      bool rw_trans= (xs->rm_error != ER_XA_RBROLLBACK);
-
-      if (rw_trans && thd->is_read_only_ctx())
-      {
-        my_error(ER_OPTION_PREVENTS_STATEMENT, MYF(0), "--read-only");
-        res= 1;
-        goto _end_external_xid;
-      }
-
       MDL_REQUEST_INIT(&mdl_request, MDL_key::BACKUP, "", "", MDL_BACKUP_COMMIT,
                        MDL_EXPLICIT);
       if (thd->mdl_context.acquire_lock(&mdl_request,
@@ -849,11 +822,7 @@ bool trans_xa_rollback(THD *thd)
     DBUG_RETURN(thd->get_stmt_da()->is_error());
   }
 
-  if (thd->transaction->all.is_trx_read_write() && thd->is_read_only_ctx())
-  {
-    my_error(ER_OPTION_PREVENTS_STATEMENT, MYF(0), "--read-only");
-    DBUG_RETURN(TRUE);
-  } else if (xid_state.xid_cache_element->xa_state == XA_ACTIVE)
+  if (xid_state.xid_cache_element->xa_state == XA_ACTIVE)
   {
     xid_state.er_xaer_rmfail();
     DBUG_RETURN(TRUE);
@@ -1136,7 +1105,7 @@ bool mysql_xa_recover(THD *thd)
 
 static bool slave_applier_reset_xa_trans(THD *thd)
 {
-  thd->variables.option_bits&= ~(OPTION_BEGIN | OPTION_KEEP_LOG);
+  thd->variables.option_bits&= ~(OPTION_BEGIN | OPTION_BINLOG_THIS_TRX);
   thd->server_status&=
     ~(SERVER_STATUS_IN_TRANS | SERVER_STATUS_IN_TRANS_READONLY);
   DBUG_PRINT("info", ("clearing SERVER_STATUS_IN_TRANS"));
